# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor
from torch import isnan,tensor
from torch import sum as summer
from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class SchweizerSklarTNorm(BaseTNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 0.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			self.p = nn.Parameter(torch.tensor(default_p))
		assert len(self.p.shape) == 0

		self.relu = torch.nn.ReLU()
		self.eps = 0.001

	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:
		res: Optional[Tensor] = None
		# Careful that the value of parametr p can be 10e-8, thus maybe add eps.
		self.adjust_nan_param()
		self.p.data = self.relu(self.p.data)
		if self.p == 0:
			self.p.data += self.eps


		res = (a ** self.p + b ** self.p) - 1.0
		res = torch.maximum(res, torch.tensor(0.0))
		res = res ** (1.0 / self.p)


		if torch.sum(torch.isnan(res)):
			print(self.p)
			print(res)
			print("GANDON NAN")

		return res

	def adjust_nan_param(self):
		for param in self.parameters():
			if summer(isnan(param)):
				param.data = tensor(0.1, device = param.device)



# -*- coding: utf-8 -*-

from torchnorms.tconorms.base import *
from torchnorms.tconorms.classic import *
from torchnorms.tconorms.dombi import *
from torchnorms.tconorms.dubois import *
from torchnorms.tconorms.hamacher import *
from torchnorms.tconorms.ss import *
from torchnorms.tconorms.weber import *
from torchnorms.tconorms.yager import *

__all__ =[
    'BaseTCoNorm',
    'BoundedTCoNorm',
    'DombiTCoNorm',
    'DuboisTCoNorm',
    'EinsteinTCoNorm',
    'HamacherTCoNorm',
    'LukasiewiczTCoNorm',
    'MinimumCoNorm',
    'NilpotentMinimumTCoNorm',
    'ProductTCoNorm',
    'SchweizerSklarTCoNorm',
    'WeberTCoNorm',
    'YagerTCoNorm',
]



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor
from torch import isnan,tensor
from torch import sum as summer
from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class HamacherTNorm(BaseTNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 1.0) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			self.p = nn.Parameter(torch.tensor(default_p))
		assert len(self.p.shape) == 0

	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:

		self.adjust_nan_param()

		res = (a * b) / (self.p + (1.0 - self.p) * (a + b - a * b))
		return res

	def adjust_nan_param(self):
		for param in self.parameters():
			if summer(isnan(param)):
				param.data = tensor(0.1, device = param.device)



# -*- coding: utf-8 -*-

import torch
from typing import Optional
from torch import nn, Tensor
from torch import isnan,tensor
from torch import sum as summer
from torchnorms.negations.base import BaseNegation


class WeberNegation(BaseNegation):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 0.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			assert default_p > -1
			self.p = nn.Parameter(torch.tensor(default_p))
		assert len(self.p.shape) == 0
		self.relu = torch.nn.ReLU()
		self.eps = 0.01
		self.__name__ = 'Weber'

	def __call__(self,
				 a: Tensor) -> Tensor:

		self.adjust_nan_param()

		self.p.data = self.relu(self.p.data + 1) - 1

		if self.p == -1:
			self.p.data += self.eps
		res = (1.0 - a) / (1.0 + self.p * a)
		return res

	def adjust_nan_param(self):
		for param in self.parameters():
			if summer(isnan(param)):
				param.data = tensor(0.1, device = param.device)



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor
from torch import isnan,tensor
from torch import sum as summer
from torchnorms.tnorms.base import BaseTNorm

from typing import Optional


class WeberTNorm(BaseTNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = -0.8) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			assert default_p >= -1.0
			self.p = nn.Parameter(torch.tensor(default_p))
		else:
			assert self.p >= -1.0

		assert len(self.p.shape) == 0

		self.relu = torch.nn.ReLU()


	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:
		res: Optional[Tensor] = None

		self.adjust_nan_param()

		self.p.data = self.relu(self.p.data + 1) - 1

		p_1 = (1.0 + self.p) * ( a + b - 1.0)
		p_2 = p_1 - self.p * a * b
		res = torch.maximum(p_2, torch.tensor(0))


		return res
		
	def adjust_nan_param(self):
		for param in self.parameters():
			if summer(isnan(param)):
				param.data = tensor(0.1, device = param.device)



# -*- coding: utf-8 -*-

import torch
from torch import nn, Tensor
from torch import isnan,tensor
from torch import sum as summer
from torchnorms.tconorms.base import BaseTCoNorm

from typing import Optional


class SchweizerSklarTCoNorm(BaseTCoNorm):
	def __init__(self,
				 p: Optional[Tensor] = None,
				 default_p: float = 0.1) -> None:
		super().__init__()
		self.p = p
		if self.p is None:
			self.p = nn.Parameter(torch.tensor(default_p))

		assert len(self.p.shape) == 0

		self.eps = 0.001

	def __call__(self,
				 a: Tensor,
				 b: Tensor) -> Tensor:

			self.adjust_nan_param()

			if self.p == 0:
				self.p.data += self.eps

			p_1 = torch.pow(1.0 - a, self.p)
			p_2 = torch.pow(1.0 - b, self.p)
			p_1 = torch.maximum(torch.tensor(0), p_1 + p_2 - 1.0)
			res = torch.pow(p_1, 1.0 / self.p)
			res = 1.0 - res

			return res

	def adjust_nan_param(self):
		for param in self.parameters():
			if summer(isnan(param)):
				param.data = tensor(0.1, device = param.device)
